package klout4j;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLEncoder;
import java.net.HttpURLConnection;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import org.json.simple.parser.JSONParser;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class Klout {
    private final String apiKey;
    private final String rootUrl;
    private final int connectTimeout;
    private final int readTimeout;
    private final String userAgent;

    private final Logger logger = LoggerFactory.getLogger(getClass());
    
    public static class Builder {
        private String apiKey;
        private String rootUrl = "http://api.klout.com";
        private int connectTimeout = -1;
        private int readTimeout = -1;
        private String userAgent = "klout4j/0.5";

        public void setApiKey(String apiKey) {
            this.apiKey = apiKey;
        }

        public void setRootUrl(String rootUrl) {
            this.rootUrl = rootUrl;
        }
        
        public void setConnectTimeout(int connectTimeout) {
            this.connectTimeout = connectTimeout;
        }

        public void setReadTimeout(int readTimeout) {
            this.readTimeout = readTimeout;
        }

        public void setUserAgent(String userAgent) {
            this.userAgent = userAgent;
        }
        
        public Klout build() {
        	return new Klout(apiKey, rootUrl, connectTimeout, readTimeout, userAgent);
        }
    }
    
    public Klout(String apiKey) {
        this.apiKey = apiKey;
        rootUrl = "http://api.klout.com";
        connectTimeout = -1;
        readTimeout = -1;
        userAgent = "klout4j/0.5";
    }
    
    private Klout(String apiKey, String rootUrl, int connectTimeout, int readTimeout, String userAgent) {
    	this.apiKey = apiKey;
    	this.rootUrl = rootUrl;
    	this.connectTimeout = connectTimeout;
    	this.readTimeout = readTimeout;
    	this.userAgent = userAgent;
    }
    
    public String getApiKey() {
    	return apiKey;
    }

    public String getRootUrl() {
        return rootUrl;
    }

    public int getConnectTimeout() {
        return connectTimeout;
    }

    public int getReadTimeout() {
        return readTimeout;
    }

    public String getUserAgent() {
        return userAgent;
    }
    
    public KloutScore getKlout(String twitterScreenName) throws KloutException {
    	List<String> twitterScreenNames = new ArrayList<String>(1);
    	twitterScreenNames.add(twitterScreenName);
    	Map<String, KloutScore> scores = getKlout(twitterScreenNames);
    	if (scores.isEmpty()) {
    		throw new ProfileNotFoundException(twitterScreenName);
    	} else {
    		return scores.entrySet().iterator().next().getValue();
    	}
    }

    public Map<String, KloutScore> getKlout(List<String> twitterScreenNames) throws KloutException {
    	logger.trace("Getting klout: " + twitterScreenNames);
        StringBuffer bld = new StringBuffer()
        		.append(rootUrl)
        		.append("/1/klout.json")
        		.append("?");
        appendKeyParameter(bld);
        bld.append("&");
        appendUsersParameter(bld, twitterScreenNames);

        if (logger.isTraceEnabled()) {
        	logger.trace("Hitting URL: " + bld.toString());
        }
        
        URL url;
        try {
        	url = new URL(bld.toString());
        } catch (MalformedURLException mue) {
        	throw new KloutException("Failed to construct URL", mue);
        }
        
        HttpURLConnection conn;
        try {
        	conn = getHttpURLConnection(url, "GET", null, false);
        	if (conn.getResponseCode() != 200) {
        		throw new KloutException(String.format(
        				"Failed to get profile detail for \"%s\", HTTP status: %d %s",
        				twitterScreenNames, conn.getResponseCode(), conn.getResponseMessage()));
        	}
        } catch (IOException ioe) {
        	throw new KloutException("Failed to open connection", ioe);
        }
        
        Map<?, ?> attributes;
        InputStream inputStream = null;
        try {
        	inputStream = conn.getInputStream();
        	attributes = (Map<?, ?>) new JSONParser().parse(new InputStreamReader(inputStream));
        } catch (Exception e) {
        	throw new KloutException("Failed to read/parse response", e);
        } finally {
        	if (inputStream != null) {
        		try {
        			inputStream.close();
        		} catch (Exception e) {
        			logger.error("Failed to close HTTP InputStream", e);
        		}
        	}
        }
        
        Object status = attributes.get("status");
        if (status == null) {
        	throw new KloutException("No status element found");
        } else if (!"200".equals(status.toString())) {
        	throw new ProfileNotFoundException(String.format(
        			"Profile for %s not found, status=%s (%s)",
        			twitterScreenNames, status, attributes.get("status_message")));
        }
        
        Map<String, KloutScore> scoresByTwitterScreenName =
        		new LinkedHashMap<String, KloutScore>();
        @SuppressWarnings("unchecked")
        List<Map<?,?>> userAttrList = (List<Map<?, ?>>) attributes.get("users");
        if (userAttrList != null) {
        	for (Map<?, ?> userAttrs : userAttrList) {
        		KloutScore score = new KloutScore(userAttrs);
        		scoresByTwitterScreenName.put(score.getTwitterScreenName(), score);
        	}
        }
        return scoresByTwitterScreenName;
    }
    
    public KloutUser showUser(String twitterScreenName) throws KloutException {
        List<String> twitterScreenNames = new ArrayList<String>();
        twitterScreenNames.add(twitterScreenName);
        Map<String, KloutUser> users = showUsers(twitterScreenNames);
        if (users.isEmpty()) {
            throw new ProfileNotFoundException(twitterScreenName);
        } else {
            return users.entrySet().iterator().next().getValue();
        }
    }

    public Map<String,KloutUser> showUsers(List<String> twitterScreenNames) throws KloutException {
        logger.trace("Showing users: " + twitterScreenNames);
        StringBuffer bld = new StringBuffer()
                .append(rootUrl)
                .append("/1/users/show.json")
                .append("?");
        appendKeyParameter(bld);
        bld.append("&");
        appendUsersParameter(bld, twitterScreenNames);

        if (logger.isTraceEnabled()) {
            logger.trace("Hitting URL: " + bld.toString());
        }
        
        URL url;
        try {
            url = new URL(bld.toString());
        } catch (MalformedURLException e) {
            throw new KloutException("Failed to construct URL", e);
        }
        
        HttpURLConnection conn;
        try {
            conn = getHttpURLConnection(url, "GET", null, false);
            if (conn.getResponseCode() != 200) {
                throw new KloutException("Failed to get profile detail for \"" +
                                         twitterScreenNames +
                                         "\", HTTP status: " +
                                         conn.getResponseCode() + " " +
                                         conn.getResponseMessage());
            }
        } catch (IOException e) {
            throw new KloutException("Failed to open connection", e);
        }

        Map<?, ?> attributes;
        InputStream inputStream = null;
        try {
            inputStream = conn.getInputStream();
            attributes = (Map<?,?>) new JSONParser().parse(new InputStreamReader(inputStream));
        } catch (Exception e) {
            throw new KloutException("Failed to read/parse response", e);
        } finally {
            if (inputStream != null) {
                try {
                    inputStream.close();
                } catch (Exception e) {
                    logger.error("Failed to close HTTP InputStream", e);
                }
            }
        }
        
        Object status = attributes.get("status");
        if (status == null) {
            throw new KloutException("No status element found");
        } else if (!"200".equals(status.toString())) {
            throw new ProfileNotFoundException("Profile for " +
                                               twitterScreenNames +
                                               " not found, status=" +
                                               status + " (" +
                                               attributes.get("status_message") +
                                               ")");
        }

        Map<String,KloutUser> usersByTwitterScreenName =
                new LinkedHashMap<String,KloutUser>();
        @SuppressWarnings("unchecked")
		List<Map<?,?>> userAttrList = (List<Map<?,?>>) attributes.get("users");
        if (userAttrList != null) {
            for (Map<?, ?> userAttrs : userAttrList) {
                KloutUser user = new KloutUser(userAttrs);
                usersByTwitterScreenName.put(user.getTwitterScreenName(), user);
            }
        }
        return usersByTwitterScreenName;
    }

    private HttpURLConnection getHttpURLConnection(URL url,
                                                   String requestMethod,
                                                   Map<String,String> headers,
                                                   boolean doOutput)
        throws java.io.IOException {

        HttpURLConnection conn = (HttpURLConnection)url.openConnection();
        if (connectTimeout != -1) {
            conn.setConnectTimeout(connectTimeout);
        }
        if (readTimeout != -1) {
            conn.setReadTimeout(readTimeout);
        }
        conn.setAllowUserInteraction(false);
        conn.setRequestMethod(requestMethod);
        conn.setDoInput(true);
        conn.setDoOutput(doOutput);
        conn.setRequestProperty("Connection", "close");
        conn.setRequestProperty("User-Agent", userAgent);
        if (headers != null) {
            for (Map.Entry<String,String> header : headers.entrySet()) {
                conn.setRequestProperty(header.getKey(), header.getValue());
            }
        }
        return conn;
    }
    
    private StringBuffer appendKeyParameter(StringBuffer buf) {
    	try {
        	buf.append("key=")
    			.append(URLEncoder.encode(apiKey, "utf-8"));
    	} catch (UnsupportedEncodingException e) {
    		throw new RuntimeException(e);
    	}
    	return buf;
    }
    
    private StringBuffer appendUsersParameter(StringBuffer buf, List<String> twitterScreenNames) {
    	buf.append("&users=");
    	int count = 0;
    	for (String twitterScreenName : twitterScreenNames) {
    		try {
	    		if (count++ > 0) {
	    			buf.append(",");
	    		}
	    		buf.append(URLEncoder.encode(twitterScreenName, "utf-8"));
    		} catch(UnsupportedEncodingException uee) {
    			throw new RuntimeException(uee);
    		}
    	}
    	return buf;
    }
}
